import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/******************************************************************************/
/* class Pile                                                                 */
/*                                                                            */
/* Abstract superclass of all the different types of pile of card.            */
/******************************************************************************/
public abstract class Pile extends Component
{
  static final int X_OFFSET = 20;
  static final int Y_OFFSET = 15;

  Vector cards = new Vector();
 
  // A pile knows about the deck's back, and how to represent an empty pile
  static Image   back    = null;
  static Image   empty   = null;  
  static Tableau tableau = null;
  
  /****************************************************************************/
  /* Constructors                                                             */
  /****************************************************************************/
  Pile(Vector v)
  {
    this();
    cards = v;
  }
  
  Pile()
  {
    if (back == null || empty == null || tableau == null)
    {
      System.err.println("Pile class static members not initialised.");
      System.exit(42);
    }
    
    final Solitaire applet = tableau.getApplet();
    
    final Pile pile = this;
      
    /**************************************************************************/
    /* All Piles have MouseListeners which check for a mouse down, mouse up   */
    /* and mouse drag.                                                        */
    /**************************************************************************/
    this.addMouseListener(new MouseAdapter() 
    {
      public void mousePressed(MouseEvent e) 
      {  
        // Mouse is pressed. Try and get a float from this pile.
        int x = e.getX();
        int y = e.getY();
        Point offset = new Point(x, y);
        Vector floated = getFloated(x, y, offset);
        
        if (floated != null)
        {
          // We've got a valid float, so tell the tableau.
          Floater floater = tableau.getFloater();
          floater.setFloat(floated, pile, offset); 
          floater.setPosition(calcPosition(e));
          tableau.repaint();       
        }
      }
      
      public void mouseReleased(MouseEvent e) 
      {  
        // Mouse is released. Unfortunately, the MouseUp goes to the same
        // pile as the MouseDown, rather than the one the mouse is over.
        // So, we have to reroute events using findComponentAt().
        Floater floater = tableau.getFloater();
        
        if (floater.isEmpty())
        {
          repaint();
          return;
        }
        
        Component comp = null;
        
        // Check for double clicks - special command.
        if (e.getClickCount() == 2)
        {
          // It's a double click, and the top card is still in the floater.
          // So, we need to attempt to put the top card on the correctly-
          // suited Foundation.
          comp = tableau.getFoundation(floater.bottom().getSuit());
        }
        else
        {
          // Simple drag; find the component we've dragged over.
          comp = applet.findComponentAt(floater.loc); 
        }
        
  	    // Try and accept cards from the floater, if there is one.
        if ((comp instanceof Pile) && 
            (comp != floater.getCreator()) && 
            (((Pile)comp).floatIsLegal()))
        {   
          ((Pile)comp).push(floater.getFloat());
        }           
        else          
        {
          // Can't accept cards from the floater.
          push(floater.getFloat());
        }
                
        tableau.repaint();       
      }      
    });
    
    this.addMouseMotionListener(new MouseMotionAdapter()
    {
      public void mouseDragged(MouseEvent e)
      {
        // Mouse drag. We need to calculate the coordinates of the floater's
        // drag position and set it.
        tableau.getFloater().setPosition(calcPosition(e));
        tableau.repaint();
      }
    });       
    
    setVisible(true);
  }  
  
  /****************************************************************************/
  /* This works out the position of the MouseEvent relative to the top left   */
  /* hand corner of the tableau.                                              */
  /****************************************************************************/
  Point calcPosition(MouseEvent e)
  {
    Point p_loc = getLocationOnScreen();
    Point t_loc = tableau.getLocationOnScreen();

    p_loc.x -= t_loc.x;
    p_loc.y -= t_loc.y;
    
    if (e != null)
    {
      p_loc.x += e.getX();
      p_loc.y += e.getY();
    }
    
    return p_loc;        
  }
  
  /****************************************************************************/
  /* All these must be overridden by subclasses:                              */
  /****************************************************************************/
  
  /****************************************************************************/
  /* Should return null if clicking at (x, y) does not produce a float, or    */
  /* the float otherwise.                                                     */
  /* The offset param is filled with the offset into the card of the click.   */
  /* This makes the drag and drop work nicely.                                */
  /****************************************************************************/
  abstract Vector getFloated(int x, int y, Point offset);
  
  /****************************************************************************/
  /* Lets a pile say what sort of floats it will accept.                      */
  /****************************************************************************/
  abstract boolean floatIsLegal();
    
  /****************************************************************************/
  /* If these aren't implemented, the Pile doesn't show, so we abstract them  */
  /* to make sure they are.                                                   */
  /****************************************************************************/
  abstract public Dimension getPreferredSize();
  abstract public void paint(Graphics g);

  /****************************************************************************/
  /* Methods implementing a stack-like interface. However, instead of pushing */
  /* and popping single items, these allow you to push and pop entire Vectors */
  /* of cards.                                                                */
  /****************************************************************************/
  boolean push(Vector v)
  {
    while (!v.isEmpty())
    {
      Card c = (Card)v.firstElement();
      cards.addElement(c);
      v.removeElement(c);
    }
    
    repaint();
    
    return true;
  }
  
  // With no arguments, we pop the entire stack of cards.
  Vector pop()
  {
    return pop(count());
  }
  
  // Pop a number of cards.
  Vector pop(int amount)
  {
    Vector retval = new Vector();
    
    while (!isEmpty() && amount > 0)
    {
      Card c = top();
      cards.removeElement(c);
      retval.addElement(c);
      
      amount--;      
    }
    
    return retval;
  }
   
  Card top() { return (Card)cards.lastElement(); }  
  
  boolean isEmpty() { return cards.isEmpty(); }  

  /****************************************************************************/
  /* count(). We can't call this size() because Component already has a       */
  /* method of that name and Java complains.                                  */
  /****************************************************************************/
  int count() { return cards.size(); }
    
  /****************************************************************************/
  /* Enable initialisation of static members with dynamic info.               */
  /****************************************************************************/
  static void setBackImage(Image i)  { back = i; }
  
  static void setEmptyImage(Image i) { empty = i; }

  static void setTableau(Tableau t) { tableau = t; }
}
