import java.applet.Applet;
import java.awt.*;
import java.util.Vector;
import java.awt.event.*;
 
/******************************************************************************/
/* class Tableau                                                              */
/*                                                                            */
/* This class is the "green baize" on which all the piles of cards sit.       */
/******************************************************************************/
public class Tableau extends Panel implements Runnable
{
  Solitaire    applet;
  
  // Piles of cards
  Talon        talon;
  Waste        waste;
  Column[]     columns = new Column[7];
  Foundation[] foundations = new Foundation[4];
  Pack         pack;
  Floater      floater = new Floater();

  // Double-buffering  
  Graphics     bufferGraphics;
  Image        bufferImage;
  
  // Card-bouncing completion animation
  boolean      completed = false;
  Thread       bouncer;
  
  /****************************************************************************/
  /* Constructor                                                              */
  /****************************************************************************/
  Tableau(Solitaire applet)
  {
    this.applet = applet;
    
    // Initialise static member of Pile class. This saves having to
    // construct every subclass with a Tableau.
    Pile.setTableau(this);

    int i;
        
    Dimension appletSize = applet.getSize();
    this.setSize(appletSize);
    
    setBackground(Color.green.darker());        
    
    // Set up the double-buffering
    bufferImage = applet.createImage(getSize().width, getSize().height);    
    bufferGraphics = bufferImage.getGraphics();
    
    // Shuffle the pack.
    pack = new Pack(this);    
    pack.shuffle();
    
    // Create a waste and a talon.
    LightweightPanel p_wt = new LightweightPanel();
    p_wt.setLayout(new FlowLayout(FlowLayout.LEFT));
    
    waste = new Waste();
    talon = new Talon(pack.deal(24), waste); 
    
    p_wt.add(talon);
    p_wt.add(waste);
    
    // Create four empty foundations.
    LightweightPanel p_fnd = new LightweightPanel();
    p_fnd.setLayout(new FlowLayout(FlowLayout.RIGHT));
    
    for (i = 0; i < 4; i++)
    {
      foundations[i] = new Foundation();
      p_fnd.add(foundations[i]);
    }
    
    // Put the top half together.
    LightweightPanel p_top = new LightweightPanel();
    p_top.setLayout(new BorderLayout());
    p_top.add("West", p_wt);
    p_top.add("East", p_fnd);

    // Create seven columns.
    LightweightPanel p_col = new LightweightPanel();
    p_col.setLayout(new FlowLayout());
        
    for (i = 0; i < 7; i++)
    {
      columns[i] = new Column(pack.deal(i + 1));
      p_col.add(columns[i]);
    } 
    
    // Assemble the entire applet.        
    setLayout(new BorderLayout());
    
    add("North", p_top);
    add("South", p_col);
    
    setVisible(true);
    
    repaint();
  }
  
  Solitaire getApplet() { return applet; }
  Floater getFloater() { return floater; }
  
  /****************************************************************************/
  /* Get the Foundation associated with a particular suit.                    */
  /****************************************************************************/
  Foundation getFoundation(Suit s)
  {
    Foundation retval = null;
    Foundation empty  = null;
    
    // Search for a foundation which matches the suit.
    for (int i = 3; i >= 0; i--)
    {
      if (foundations[i].count() > 0) 
      {
        if(foundations[i].top().getSuit() == s)        
        {
          retval = foundations[i];
        }
      }      
      else
      {
        empty = foundations[i];
      }
    }
    
    // Return the first empty Foundation if we don't have one of that suit.
    if (retval == null)
    {
      retval = empty;
    }
    
    return retval;
  }
   
  /****************************************************************************/
  /* Check to see if the game is finished. Called by Foundation.push().       */
  /****************************************************************************/
  void checkForWin()
  {
    int i;

    for (i = 0; i < 4; i++)
    {
      if (foundations[i].count() < 13)
      {
        return;
      }
    }
    
    // We need to repaint the last Foundation that got dragged off.
    // Hopefully this will complete before the paint() method changes.
    repaint();
    
    bouncer = new Thread(this);

    completed = true;        
    bouncer.start();    
    
  }
  
  /****************************************************************************/
  /* To do the bouncing card thing, we fire off the bouncer as a separate     */
  /* thread and let the main thread do the painting.                          */
  /* The bouncing effect comes for free - we just make updates to the         */
  /* double buffer increment-only (see paint()) and then turn each card in    */
  /* turn into a floater, and bounce it.                                      */
  /****************************************************************************/
  public void run()
  {
    int j;

    // Stop if we're being closed.
    if (!completed)
    {
      return;
    }
    
    // There are 13 cards on four foundations. We take one from each
    // foundation in turn.    
    int height = getBounds().height - Card.HEIGHT;
    int width = getBounds().width;

    for (int i = 0; i < 13; i++)
    {
      for (j = 0; j < 4; j++)
      {
        // Get the foundation.
        Foundation f = foundations[j];    
        Point start_pos = f.calcPosition(null);
        Point current_pos;

        // Set the card up as a floater.
        floater.setFloat(f.pop(1), f, new Point(0, 0));
        floater.setPosition(start_pos); 
        current_pos = new Point(start_pos);
        
        // Set up the physics with a random element
        double x_vel = 20 - (Math.random() * 40);      /* 20   to -20  */
        
        // Don't allow it to bounce up and down on the spot -
        // instead, fire it off hard to the left.
        if (x_vel > -2 && x_vel < 2)
        {
          x_vel = -20 - (Math.random() * 10);          /* -20  to -30  */
        }

        double y_vel = -10 + (Math.random() * 20);     /* -10  to +10  */
        double y_vel_damp = (Math.random() + 1.9) / 3; /* 0.96 to 0.63 */
        double y_accel = 2;

        // Let's go...
        while ((current_pos.x > -Card.WIDTH) && 
               (current_pos.x < width))
        {
          // Simple motion
          y_vel = y_vel + y_accel;
          current_pos.x = current_pos.x + (int)x_vel;
          current_pos.y = current_pos.y + (int)y_vel;
          
          if (current_pos.y >= height)
          {
            // Bounce :-)
            current_pos.y = height;
            y_vel = (-y_vel) * y_vel_damp;
          }

          // Move the floater and ask for a repaint...
          floater.setPosition(current_pos);
          floater.draw(bufferGraphics);
          repaint();

          // Then sleep so it can happen.
          try
          {
            bouncer.sleep(50);
          }
          catch (InterruptedException e)
          {
          }
        }
      }      
    }
  } 
   
  /****************************************************************************/
  /* Painting                                                                 */
  /****************************************************************************/
  public boolean isDoubleBuffered() { return true; }  
 
  public void paint(Graphics g)
  {
    // If we haven't finished, we update the double buffer with the state of
    // the tableau and then draw the floater on top. This could be improved
    // by only redrawing changed bits.
    if (!completed)
    {
      bufferGraphics.setColor(getBackground());
      bufferGraphics.fillRect(0, 0, getBounds().width, getBounds().height);
      
      super.paint(bufferGraphics);
      floater.draw(bufferGraphics); 
       
      // Copy to the Graphics
      g.drawImage(bufferImage, 0, 0, null);        
    }
    else
    {
      // If we've succeeded, we are bouncing, so just draw the current state
      // of the image. The floater will be drawing to the Image in a separate 
      // thread.
      g.drawImage(bufferImage, 0, 0, null);        
    }    
  }
  
  public void update(Graphics g)
  {
    paint(g);
  } 
}
